/*
 * (./) Track.java 05/11/14 by cs for "MMM" project
 * (cc) some right reserved
 */

package mmm;

import java.awt.Color;
import java.util.Vector;


/* a simple object representing a track element
 */
public class Track {
	
	// define the interval (in second) between two different output
	public static final double SIGNAL_INTERVAL = 3.0;
	// define the 'clickeable' entry area
	public static final float ENTRY_TOLERANCE = 7f;
	
	// define the Output entries values
	public static final int TYPE_IN		 = 1;
	public static final int TYPE_OUT	 = 0;
	public static final int TYPE_BOTH	 = 2;	// both entries
	public static final int TYPE_NULL	 = -1;	// no entries available
	
	// declare properties
	public int index;
	public boolean enable;
	public boolean visible;
	public Color color;
	public String name;
	
	
	// store all outputs
	private Vector outputs	= new Vector();
	
	// use for moving signals: the selected signal, the selection type, the
	// differences between the clicked time point and the entries values, 
	// the selection minimum and maximum entries values
	private Output select	 = null;
	private int entry		 = TYPE_NULL;
	private double delta_in	 = .0;
	private double delta_out = .0;
	private double min		 = .0;
	private double max		 = -1.;	// means infinity
	
	// the track duration
	private double length	= .0;	
	
	// an output (a signal) is an object construct with two properties
	// representing the start or stop time values in seconds. To retrieve the 
	// correct maximum or minimum value call the in() or out() method:
	// in() <= out()
	class Output {
		
		private double s1, s2; 
		Output( double in, double out ) { in(in); out(out); }
		
		// set the first property but get the real minimum value
		void in( double S ) { s1=S; }
		double in() { return s1<=s2 ? s1 : s2; }
		
		// set the second property but get the real maximum value
		void out( double S ) { s2=S; }
		double out() { return s2>s1 ? s2 : s1; }
		
	}
	
	
	/* constructors
	 * @arg int, the track index
	 * @arg boolean, the track enable status
	 * @arg boolean, the track visible status
	 * @arg Color, the track color
	 * @arg String, the track name
	 */
	public Track() {;}
	public Track
	( int idx, boolean isEnabled, boolean isVisible, Color color, String name )
	{
		this.index		= idx;
		this.enable		= isEnabled;
		this.visible	= isVisible;
		this.color		= color;
		this.name		= name;
	}
	
	/* set track properties
	 */
	public void setIndex( int i )		{ index	= i; }
	public void setEnabled( boolean b )	{ enable = b; }
	public void setVisible( boolean b )	{ visible = b; }
	public void setColor( Color c )		{ color	= c; }
	public void setName( String s )		{ name = s; }
	
	/* get track properties
	 */
	public int getIndex()		{ return index; }
	public boolean getEnabled() { return enable; }
	public boolean getVisible() { return visible; }
	public Color getColor()		{ return color; }
	public String getName()		{ return name; }
	
	/* convert each properties intop Object
	 * and return the result as Object list
	 * @arg int, the track index
	 */
	public Object[] toObject() {
		return new Object[]{
			new Integer( index ),
			new Boolean( enable ),
			new Boolean( visible ),
			color,
			name
		};
	}
	
	/* create and store a new output.
	 * Set it as the current selected signal.
	 */
	public void add( double s ) { add(s,s); }
	public void add( double in, double out ) {
		outputs.add( select = new Output( in, out ) );
		entry	= TYPE_OUT;
		setRange();
	}
	
	/* remove selected signals
	 */
	public boolean remove() {
		boolean result = false;
		if ( select!=null && outputs.remove(select) ) {
			select = null;
			result = true;
		}
		return result;
	}
	
	/* set the selected output according to the selected entry
	 * @arg double, the required output value
	 */
	public void set( double d ) {
		if ( select==null ) return;
		
		double in  = getIn();
		double out = getOut();
		
		// switch entry if needed
		if ( entry!=TYPE_BOTH && d<=in ) entry = TYPE_IN;
		if ( entry!=TYPE_BOTH && d>out ) entry = TYPE_OUT;
		
		switch( entry ) {
			case TYPE_IN	: select.in( Math.max(d, min) ); break;
			case TYPE_OUT	: select.out( Math.min(d, max) ); break;
			case TYPE_BOTH	:
				// reevaluate in/out value
				double len = delta_in+delta_out;
				in  = Math.max( Math.min( max-len, d-delta_in ),  min );
				out = Math.min( Math.max( min+len, d+delta_out ), max );
				select.in( in );
				select.out( out );
				break;
		}
	}
	
	/* get the minimum or maximum output value
	 */
	public double getIn()	{ return (select==null) ? 0. : select.in(); }
	public double getOut()	{ return (select==null) ? 0. : select.out(); }
	
	/* set the track duration = the maximum value in seconds
	 */
	public void setLength( double len )	{ length = len; }
	
	/* define the selection range :
	 * set the minimum and the maximum values for the selected entries
	 */
	private void setRange() {
		if ( select==null ) return;
		
		min = .0;
		max = length;
		
		for( int i=0; i<outputs.size(); i++ ) {
			Output o = (Output)outputs.get(i);
			if ( o==select ) continue;
			if ( select.in()>=o.out() && o.out()>min ) min = o.out();
			if ( select.out()<=o.in() && o.in()<max  ) max = o.in();
		}
	}
	
	/* get the minimum and maximum values for the selected entries
	 */
	public double getMin()	{ return min; }
	public double getMax()	{ return max; }
	
	/* search and select the required signal at the given position, and by 
	 * extension the output's entry.
	 * return true on success, false otherwise.
	 * @arg double, the time position value in second
	 */
	public boolean select( double s ) {
		for( int i=0; i<outputs.size(); i++ ) {
			Output o = (Output)outputs.get(i);
			if ( (o.in()-SIGNAL_INTERVAL)<=s && s<=(o.out()+SIGNAL_INTERVAL) ) {
			// signal is found
				entry	= TYPE_BOTH; // if the given time points to no entries
				// register the actual differences between the clicked time point
				// and the in/out values
				delta_in	= s-getIn();
				delta_out	= getOut()-s;
				// set the current entry
				if ( s<=(o.in()+ENTRY_TOLERANCE) )	entry = TYPE_IN;
				if ( s>=(o.out()-ENTRY_TOLERANCE) )	entry = TYPE_OUT;
				// select this signal
				select = o;
				// set the minimum and maximum value
				setRange();
				// change track listening status and return
				return true;
			}
		}
		return false; // no signal
	}
	
	/* deselect the selected output
	 */
	public void deselect() {
		select	= null;
		entry	= TYPE_NULL;
	}
	
	/* get all in/out values as a double list
	 * i.e [ in0, out0, in1, out1, ... inX, outX ]
	 */
	public double[] getOutputs() {
		
		// store all outputs values
		double[] values = new double[ outputs.size()*2 ];
		for( int i=0; i<outputs.size(); i++ ) {
			Output output	= (Output)outputs.get(i);
			values[i*2]		= output.in();
			values[(i*2)+1]	= output.out();
		}
		
		// return it
		return values;
	}
	
	/* set/get the selected signal's entry
	 */
	public void setSelectionType( int which ) { entry = which; }
	public int getSelectionType() { return ( select==null ? TYPE_NULL:entry); }
	
	/* return the selected output index or -1
	 */
	public int getSelectionIndex() { return outputs.indexOf(select); }
	
	/* check for the track output status at the given position.
	 * return true while TYPE is set to TYPE_IN or TYPE_BOTH, false otherwise.
	 * @arg double, the time position value in second
	 */
	public boolean checkStatus( double s ) {
		if ( !enable ) return false;
		
		for(int i=0; i<outputs.size(); i++ ) {
			
			double in	= ( (Output)outputs.get(i) ).in();
			double out	= ( (Output)outputs.get(i) ).out();
			
			if ( s==in )			return true;
			if ( s==out )			return false;
			if ( s>in && s<out )	return true;
		}
		
		// default 
		return false;
	}
}
