/*
 * (./) TrackGraphic.java 05/11/14 by cs for "MMM" project
 * (cc) some right reserved
 */

package mmm.ui;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.event.MouseInputListener;

import mmm.Track;

/* a graphic component to manage manually project's tracks
 */
class TrackGraphic extends JComponent {
	
	// define track height
	protected final static int SIZE_XSMALL	= 10;
	protected final static int SIZE_SMALL	= 25;
	protected final static int SIZE_MEDIUM	= 50;
	protected final static int SIZE_LARGE	= 75;
	protected final static int SIZE_XLARGE	= 100;
	
	// graphic colors
	protected final static Color COLOR_DISEABLE			= new Color(240,240,240);
	protected final static Color COLOR_TEXT				= Color.WHITE;
	protected final static Color COLOR_SIGNAL			= Color.BLACK;
	protected final static Color COLOR_SELECTED_SIGNAL	= Color.RED;
	protected final static Color COLOR_SELECTED_TRACK	= Color.RED;
	protected final static Color COLOR_CURSOR_ON		= Color.RED;
	protected final static Color COLOR_CURSOR_OFF		= new Color(0,0,0,30);
	
	// track dimension
	private int width	= 0;
	private int height	= SIZE_MEDIUM;
	
	// paint required properties
	private Vector toBeDraw	= new Vector();	// visible graphics index
	private Vector buffer	= new Vector();	// tracks image
	
	// the owner component
	private JComponent owner = null;
	
	// the selected track
	// -1 = no selection
	private int select				= -1;
	private static Image selectIMG	= Resource.getImage( "select.png" );
	
	// the graphic cursor 
	// ( display only on play mode )
	private int cursor = -1;
	
	// true while playing
	private boolean isRunning = false;
	
	
	/* constructor
	 */
	public TrackGraphic() {;}
	
	/* set graphics width
	 * @arg int, the width value
	 */
	public void setWidth( int i ) { 
		width = i;
		setSize();
	}
	
	/* set graphics height
	 * @arg int, the width value
	 */
	public void setHeight( int i ) {
		height = i;
		setSize();
	}
	
	/* set the component's size according to the track's dimension and the
	 * number of tracks to be displayed
	 */
	public void setSize() {
		setPreferredSize(
			new Dimension( width+Editor.GRAPHIC_RIGHT_MARGIN, 
						   (height+1)*toBeDraw.size()
			) 
		);
		
		if ( owner!=null ) owner.revalidate();
	}
	
	/* set up the graphic panel owner component
	 * @arg Component, the owner component
	 */
	public void setOwner( JComponent comp ) { owner = comp; }
	
	/* append a new buffered image with the specified properties
	 * @arg Track, the track object to be represented
	 */
	public void append( Track track ) {
		if ( track.visible ) toBeDraw.add( new Integer(buffer.size()) );
		buffer.add( getImage(track) );
	}
	
	/* update an existant graphic
	 * @arg Track, the updated track object
	 */
	public void update( Track track ) {
		
		int at, i, cur;
		
		if ( !track.visible ) toBeDraw.remove( new Integer(track.index) );
		else {
			Integer index = new Integer(track.index);
			if ( toBeDraw.indexOf(index)==-1 ) {
				at = toBeDraw.size();
				for( i=0; i<toBeDraw.size(); i++ ) {
					cur = ((Integer)toBeDraw.get(i)).intValue();
					if ( cur > track.index ) {
						at=i;
						break;
					}
				}
				toBeDraw.add(at, index);
			}
		}
		buffer.set( track.index, getImage(track) );
	}
	
	/* insert a new buffered image at the track index position
	 * @arg Track, the track object to be insert
	 */
	public void insert( Track track ) {
		if ( track.visible ) {
			if ( track.index >= toBeDraw.size() )
				 toBeDraw.add( new Integer(track.index) );
			else toBeDraw.add( track.index, new Integer(track.index) );
			reassign( track.index+1, 1 );
		}
		buffer.add( track.index, getImage(track) );
	}
	
	/* remove the graphic at the specified position
	 * @arg int, the track index to be remove
	 */
	public void remove( int index ) {
		
		int from = toBeDraw.indexOf( new Integer(index) );
		
		// remove 
		toBeDraw.remove( from );
		buffer.remove( index );
		
		// update index list
		reassign( from, -1 );
	}
	
	/* reassign images indexes (to be draw)
	 * @arg int, the started index (inclusive)
	 * @arg int, the increminent value
	 */
	private void reassign( int from, int by ) {
		for( int i=from; i<toBeDraw.size(); i++ ) {
			int j = ((Integer) toBeDraw.get(i)).intValue();
			toBeDraw.set(i, new Integer(j+by));
		}
	}
	
	/* create a new image with the track properties
	 * @arg Track, the track to be draw
	 * @type Image
	 */
	public Image getImage( Track track ) {
		
		BufferedImage img = new BufferedImage( Math.max( 1, width ), 
											   height, 
											   BufferedImage.TYPE_INT_ARGB
											);
		Graphics2D gc2D	  = (Graphics2D) img.getGraphics();
		Color color		  = track.enable ? track.color : COLOR_DISEABLE;
		double[] outputs  = track.getOutputs();
		int y			  = ((height/25)*2)+9;
		
		// set renderer :
		// enable antialiasing for graphic
		// disable antialiasing for the text
        gc2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
							  RenderingHints.VALUE_ANTIALIAS_ON);
        gc2D.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, 
							  RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		
		// draw background
		gc2D.setColor( color );
		gc2D.fillRect( 0, 0, width, height );
		
		// draw track name
		gc2D.setFont( Editor.FONT_BOLD );
		gc2D.setColor( COLOR_TEXT );//color.brighter() );
		gc2D.drawString( getPrefix(track.index), 16, y );
		gc2D.setFont( Editor.FONT_PLAIN );
		gc2D.drawString( track.name, 40, y );

		// draw disable text
		if ( !track.enable ) {
			gc2D.setFont( Editor.FONT_BOLD );
			gc2D.setColor( Color.RED );
			gc2D.drawString( "X", 28, y );
		}
		
		// draw markers
		y = height/2;
		for( int i=0; i<outputs.length; i +=2 ) {
			
			int in_s	= (int)outputs[i];						// start in
			int	in_e	= (int)( in_s+Track.ENTRY_TOLERANCE );	// end in
			int out_e	= (int)outputs[i+1];					// start out
			int	out_s	= (int)( out_e-Track.ENTRY_TOLERANCE );	// end out
			
			gc2D.setColor( COLOR_SIGNAL );//color.darker() );
			// draw signal
			gc2D.drawLine( in_s, y-1, out_e, y-1);
			// draw entries
			if ( out_e>in_e ) {
				gc2D.fillPolygon( new int[]{ in_s, in_e, in_e }, 
								  new int[]{ y-1, y-4, y+2 }, 
								  3
								);
				gc2D.fillPolygon( new int[]{ out_s, out_e, out_s }, 
								  new int[]{ y-4, y-1, y+2 }, 
								  3
								);
			}
			
			
			// draw selection
			// MAY BE IMPLEMENTED IN DRAW METHOD DIRECTLY
			if ( in_s!=out_e && track.getSelectionIndex()==i/2 ) {
				gc2D.setColor( COLOR_SELECTED_SIGNAL );//color.brighter() );
				// signal
				if ( track.getSelectionType()==Track.TYPE_BOTH ) 
					gc2D.drawLine( in_e, y-1, out_s, y-1);
				// in 
				if ( track.getSelectionType()==Track.TYPE_IN || 
					 track.getSelectionType()==Track.TYPE_BOTH )
					gc2D.drawOval( in_s-1,
								   (int) (y-(Track.ENTRY_TOLERANCE/2))-1,
								   (int) (Track.ENTRY_TOLERANCE)+1,
								   (int) (Track.ENTRY_TOLERANCE)+1
							   );
				// out
				if ( track.getSelectionType()==Track.TYPE_OUT || 
					 track.getSelectionType()==Track.TYPE_BOTH )
					gc2D.drawOval( out_s-1,
								   (int) (y-(Track.ENTRY_TOLERANCE/2))-1,
								   (int) (Track.ENTRY_TOLERANCE)+1,
								   (int) (Track.ENTRY_TOLERANCE)+1
								   );
			}
		}
		
		// return the result
		return img;
	}
	
	/* get the track name
	 * @arg int, the track index
	 * @arg string, the track name
	 */
	private String getPrefix( int i ) {
		String prefix = (i+1)+" - ";
		if ( i<9 ) return " "+prefix;
		return prefix;
		/*String prefix = "";
		switch( i/6 ) {
			case 0 : prefix = "JD."; break;
			case 1 : prefix = "BD."; break;
			case 2 : prefix = "JG."; break;
			case 3 : prefix = "BG."; break;
		}
		return prefix+((i%6)+1);*/
	}
	
	/* paint method, draw all visible tracks
	 */
	public void paint( Graphics g ) {
		
		// paint only the changed tracks (ie the selected track)
		// paint only the visible graphic components area 
		// > g.getClipBounds();
		// WELL... OPTIMIZE :)
		int i=0;
		
		// draw background
		g.setColor( Color.WHITE );
		g.fillRect( 0, 0, width, (height+1)*toBeDraw.size()-1 );
		
		// draw tracks
		g.setColor( COLOR_SELECTED_TRACK );
		for( i=0; i<toBeDraw.size(); i++ ) {
			
			int y	  = (height+1)*i;
			int index = ((Integer) toBeDraw.get(i)).intValue();
			Image img = (Image) buffer.get(index);
			
			// track image
			g.drawImage( img, 0, y, null );
			// draw selection
			if ( select==index ) //g.drawRect( 0, y-1, width-1, height+1);
				g.drawImage( selectIMG, 2, y+((height/25)*2), null );
		}
		
		
		// draw the cursor line
		if ( cursor==-1 ) return;
		if ( isRunning ) {
			g.setColor( Color.PINK );
			g.fillRect( cursor-2, 0, 6, (height+1)*i );
		}
		g.setColor( isRunning ? COLOR_CURSOR_ON : COLOR_CURSOR_OFF );
		g.drawLine( cursor, 0, cursor, (height+1)*i );
	}
		
	/* set the mouse listener
	 */
	public void setMouseListener( MouseInputListener listener ) {
		addMouseListener( listener );
		addMouseMotionListener( listener );
	}
	
	/* return the track index at a specified position
	 * or -1 if no track is found
	 * @arg int, the x coordonate
	 * @arg int, the y coordonate 
	 */
	public int getTrackIndex( int x, int y ) {
		
		int index  = -1;			// index to be return
		int search = y/(height+1);	// search visible buffer index
		
		if ( toBeDraw.size()==0 || search>=toBeDraw.size() || x>width ) 
			return index;
		return ((Integer) toBeDraw.get(search)).intValue();
	}
	
	/* move the cursor
	 * @arg int, the new cursor value
	 */
	public void setCursor( int x ) { setCursor( x, false ); }
	public void setCursor( int x, boolean run ) {
		cursor		= Math.max( Math.min(x, width), 0 );
		isRunning	= run;
		repaint();
	}
	
	/* set the selected track
	 */
	public void selectTrack( int index ) { select = index; }
}

